#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['scanpydoc']

package_data = \
{'': ['*'], 'scanpydoc': ['static/*']}

install_requires = \
['sphinx>=1.8', 'get_version', 'future-fstrings']

extras_require = \
{'doc': ['sphinx-autodoc-typehints'],
 'test': ['pytest', 'pytest-cov', 'scipy'],
 'test:python_version != "3.5"': ['pytest-black']}

setup(name='scanpydoc',
      version='0.3.2',
      description='A series of Sphinx extensions to get easy to maintain, numpydoc style documentation.',
      author='Philipp Angerer',
      author_email='phil.angerer@gmail.com',
      url='https://icb-scanpydoc.readthedocs-hosted.com/en/latest/',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.5',
     )
